import os

from flask import Flask, render_template


class PlotManager:
    """
    绘图工具
    """

    def __init__(self):
        # 寻求路径
        self.BASE_DIR = os.path.dirname(os.path.abspath(__file__))
        self.templates_dir = os.path.join(self.BASE_DIR, 'templates')
        self.static_dir = os.path.join(self.BASE_DIR, 'static')

    def plot_show(self):
        # 指定 templates 和 static
        _app = Flask(__name__, template_folder=self.templates_dir, static_folder=self.static_dir)

        def plot():
            strategies = []
            return render_template('k-line.html', strategies=strategies)

        _app.add_url_rule('/', view_func=plot)
        _app.run(host='127.0.0.1', port=9999)
        return _app


if __name__ == '__main__':
    plot = PlotManager()
    plot.plot_show()



class RingBuffer:
    """ class that implements a not-yet-full buffer """

    def __init__(self, size_max):
        self.max = size_max
        self.data = []

    class __Full:
        """ class that implements a full buffer """

        def append(self, x):
            """ Append an element overwriting the oldest one. """
            self.data[self.cur] = x
            self.cur = (self.cur + 1) % self.max

        def get(self):
            """ return list of elements in correct order """
            return self.data[self.cur:] + self.data[:self.cur]

    def append(self, x):
        """append an element at the end of the buffer"""
        self.data.append(x)
        if len(self.data) == self.max:
            self.cur = 0
            # Permanently change self's class from non-full to full
            self.__class__ = self.__Full

    def get(self):
        """ Return a list of elements from the oldest to the newest. """
        return self.data


def test02():
    ll = []
    for i in range(10):
        if i > 5:
            del ll[0]
        ll.append(i)
    print(ll)


# sample usage
if __name__ == '__main__':
    # x = RingBuffer(5)
    # x.append(1)
    # x.append(2)
    # x.append(3)
    # x.append(4)
    # print(x.__class__, x.get())
    # x.append(5)
    # print(x.__class__, x.get())
    # x.append(6)
    # print(x.data, x.get())
    # x.append(7)
    # x.append(8)
    # x.append(9)
    # x.append(10)
    # print(x.data, x.get())

    test02()



from framework.log_handler import DefaultLogHandler
from framework.quote_manager import QuoteManager

####################################################
# 行情管理器使用例子
####################################################

log = DefaultLogHandler(filename='GetQuoteDemo.log')
log.info('************** 获取行情数据例子 ***************')

# 创建一个QuoteManager实例
quoteM = QuoteManager(isShowQuoteMsg=False)

# 获取历史数据，不使用回调函数，直接拿返回值
# data = quoteM.getHistory(stock_code='COMEX_F_GC_2012', startTime='2020-06-07', endTime='2020-08-17', ktype='1Min')
# print(data)


# 注意 QuoteManager的自定义回调方法不能执行耗时操作，否则会影响下一步执行
def listener_fun(data):
    print(f'1min > {data}')


def listener_fun2(data):
    print(f'5min > {data}')

# 获取历史数据，使用回调函数，历史数据会一条一条模拟行情产生，回调自定义函数
quoteM.getHistory(stock_code='COMEX_F_GC_2012', startTime='2020-08-07', endTime='2020-08-17'
                  , ktype='1Min', fun_listener=listener_fun)

# 获取实时行情数据
# quoteM.getRealTimeKL(stock_code='CFFEX_F_IF_2009', ktype='1Min', fun_listener=listener_fun)
# quoteM.getRealTimeKL(stock_code='CFFEX_F_IF_2009', ktype='5Min', fun_listener=listener_fun2)

quoteM.stop()



import json


def file2dict(path):
    """
    从文件读取配置信息转字典
    :param path: 配置文件路径
    :return:
    """
    try:
        # 读取配置文件
        with open(path) as f:
            return json.load(f)
    except Exception as e:
        print(e.args)


def select_pan_type(code: str) -> str:
    """
    判断内外盘
    :param code:NYMEX_F_CL_2007
    :return:
    """
    contratelist = str(code).split('_')
    if contratelist[0].upper() in ['SHFE', 'CFFEX', 'DCE', 'CZCE']:
        return 'in'  # 内盘
    else:
        return 'out'  # 外盘


def isKtypeRight(ktype):
    if ktype in ['1Min', '5Min', '15Min', '30Min', '60Min', 'klDay', 'klWeek']:
        return True
    return False



import traceback
from collections import defaultdict
from queue import Queue
from threading import Thread

from framework.log_handler import DefaultLogHandler


class klEvent:
    """
    K 线事件
    """

    def __init__(self, theme: str, data, is_async=True):
        """

        :param theme: 事件类型，字符串
        :param data: 数据内容，没指定类型，取出的时候要注意类型转换
        :param is_async: 是否异步,默认True,True是不阻塞其它事件执行
        """
        self.theme = theme
        self.data = data
        self.is_async = is_async


class EventEngine:
    """事件驱动引擎"""

    def __init__(self, logHandler=None):
        if logHandler is None:
            self._logHandler = DefaultLogHandler(filename='EventEngine.log')
        else:
            self._logHandler = logHandler
        """初始化事件引擎"""
        # 事件队列
        self.__queue = Queue()
        # 事件引擎开关
        self.__active = False

        # 事件引擎处理线程
        self.__thread = Thread(target=self.__run, name="EventEngine.__thread")

        # 事件字典，key 为时间， value 为对应监听事件函数的列表
        self.__handlers = defaultdict(list)
        self.__handlers_thread = []

    def __run(self):
        """启动引擎"""
        while self.__active or self.__queue.qsize() != 0:  # 执行完队列的元素才借宿
            try:
                event = self.__queue.get(block=True)
                if not self.__active or event.theme == 'exit-loop':
                    break
                # 检查该事件是否有对应的处理函数
                if event is None or event.theme not in self.__handlers:
                    continue
                if event.is_async:
                    # 一个事件一条线程处理
                    handle_thread = Thread(target=self.__process, name="event", args=(event,))
                    # self.__handlers_thread.append(handle_thread)
                    handle_thread.start()
                else:
                    self.__process(event)
            except Exception as e:
                self._logHandler.error(e.args)
                self._logHandler.error(traceback.print_exc())
        self._logHandler.debug('---> EventEngine loop stop')
        self.__active = False
        self.__thread = None

    def __process(self, event: klEvent):
        """事件处理"""
        # 若存在,则按顺序将时间传递给处理函数执行
        for handler in self.__handlers[event.theme]:
            try:
                handler(event.data)
            except Exception as e:
                self._logHandler.error(e.args)
                self._logHandler.error(traceback.print_exc())
        # TODO 回收线程资源
        # self.__handlers_thread.pop()

    def isActivate(self):
        """
        是否开启线程
        :return:
        """
        return self.__active or self.__queue.qsize() != 0

    def start(self) -> bool:
        """引擎启动"""
        self.__active = True
        if self.__thread is None:
            self.__thread = Thread(target=self.__run, name="EventEngine.__thread")
        if self.__thread.is_alive():
            self._logHandler.info('---> EventEngine 线程已存在')
            self.__thread.join()
            self.__thread = Thread(target=self.__run, name="EventEngine.__thread")
        self.__thread.start()
        self._logHandler.debug('---> EventEngine start loop.')
        return True

    def stop(self):
        """停止引擎"""
        self.__active = False
        self.__queue.put(klEvent(theme='exit-loop', data=''))
        if self.__thread is not None:
            self.__thread.join()
        self._logHandler.debug('---> EventEngine stop finished.')

    def register(self, theme, handler):
        """注册事件处理函数监听"""
        if handler not in self.__handlers[theme]:
            self.__handlers[theme].append(handler)

    def unregister(self, theme, handler):
        """注销事件处理函数"""
        handler_list = self.__handlers.get(theme)
        if handler_list is None:
            return
        if handler in handler_list:
            handler_list.remove(handler)
        if len(handler_list) == 0:
            self.__handlers.pop(theme)

    def put(self, event):
        self.__queue.put(event)

    @property
    def queue_size(self):
        return self.__queue.qsize()



from framework.log_handler import DefaultLogHandler
from framework.strategies_manager import StrategiesManager
from framework.trader.trade_manager import Direction, OffsetFlag

####################################################
# 交易管理器使用例子
# 持仓列表格式：
# [合约，最新价，持仓均价，数量，方向，开平标志，交易费，滑点，浮动盈亏，订单时间, 账户余额，描述]
#
# 订单历史记录格式：
# ['合约', '最新价', '成交价', '数量', '方向', '开平标志', '交易费', '滑点', '订单时间', '账户余额', '描述']
#
# 策略管理器跟交易管理器结合使用
####################################################

log = DefaultLogHandler(filename='TradeDemo.log')
log.info('************** 综合交易例子 ***************')

# 创建实例，参数可以不传
s_manager = StrategiesManager(configPath='./sconfig.conf', tempPath='./temp', isShowQuoteMsg=False)

trade_m = s_manager.getTradeManager()
trade_m.check_in_money(100000)

current_time = ''

# contract = 'COMEX_F_GC_2012'
# contract = 'NYMEX_F_IF_2009'
# contract = 'NYMEX_F_CL_2010'
contract = 'HKEX_F_HSI_2009'


def myAction01(msg):
    global s_manager
    global trade_m
    global log
    global contract
    log.info(msg)
    data = s_manager.get_history_quote(contract)
    if data is not None and len(data) > 2:
        global current_time
        if current_time != msg['time_key']:
            rr2 = data[-3]
            rr1 = data[-2]
            if rr1['open'] < rr1['close'] and rr2['close'] > rr2['open'] and rr1['open'] >= rr2['open']:
                """
                手续费计算分2种情况：
                （1）固定手续费
                    N手期货固定手续费 = N手*固定手续费
                （2）比例手续费
                    N手期货固定手续费 = N手*价格*交易单位*费率
                """
                trade_m.order(contract=contract
                              , money=float(msg['close'])
                              , volume=1
                              , orderDate=msg['time_key']
                              , direction=Direction.TM_BUY
                              , offset_flag=OffsetFlag.TM_Open
                              , transferfee=20 * 1
                              , slip=0)
                # log.info(f'账户余额：{trade_m.get_account_msg()}')
                log.info(f'买入：{trade_m.get_all_position()}')
        current_time = msg['time_key']


def myAction011(msg):
    global s_manager
    global trade_m
    global log
    global contract
    log.info(msg)
    data = s_manager.get_history_quote(contract)
    if data is not None and len(data) > 1:

        rr1 = data[-2]  # 前一根K线
        if rr1['open'] < rr1['close'] and msg['open'] > rr1['open']:
            trade_m.order(contract=contract
                          , money=float(msg['open'])
                          , volume=1
                          , orderDate=msg['time_key']
                          , direction=Direction.TM_BUY
                          , offset_flag=OffsetFlag.TM_Open
                          , transferfee=20 * 1
                          , slip=0)
            log.info(f'买入：{trade_m.get_all_position()}')


def myAction02(msg):
    """
    风控
    :param msg:
    msg 数据格式
    :return:
    """
    global trade_m
    global log
    global contract

    # 获取合约持仓
    position = trade_m.get_position(contract=contract
                                    , direction=Direction.TM_BUY
                                    , offset_flag=OffsetFlag.TM_Open)
    # 如果有持仓，且不是该行情时间成交的（实时行情是存在同分钟推送几次的问题，不在同一分钟内同时执行买卖操作）
    if position is not None: #  and msg['time_key'] != position[9]
        # 持仓数据格式：
        # [0 合约，1 最新价，2 持仓均价，3 数量，4 方向，5 开平标志，6 交易费，7 滑点，8 浮动盈亏，9 订单时间, 10 账户余额，11 描述]
        pre = round((float(msg['close']) - float(position[2])) / float(position[2]) * 100, 2)
        # pre = round((float(msg['close']) - float(msg['last_close'])) / float(msg['last_close']) * 100, 2)

        log.info(f'涨跌幅:{pre}')
        if pre >= 0.07 or pre < 0:  # 止盈止损，涨跌幅超过 1%
            log.info(f'################################')
            log.info(f'持仓：{position}')
            trade_m.order(contract=contract
                          , money=float(msg['close'])
                          , volume=int(position[3])
                          , orderDate=msg['time_key']
                          , direction=Direction.TM_SELL
                          , offset_flag=OffsetFlag.TM_Close
                          , transferfee=20 * int(position[3])
                          , slip=0)
            log.info(f'################################')


def myAction03(msg):
    print(msg)


# 注册自定义函数
# s_manager.registAction(func=myAction01, contract_tag=contract)
s_manager.registAction(func=myAction02, contract_tag=contract)
s_manager.registAction(func=myAction011, contract_tag=contract)

# s_manager.registAction(func=myAction03, contract_tag='COMEX_F_GC_2012')

# 实时数据使用，支持内外盘合约 CFFEX_F_IF_2009
# s_manager.runRealTime(stock_code=contract, ktype='1Min', isGetAllQuote=False)
# 历史数据回测使用，支持内外盘合约
s_manager.runHistory(stock_code=contract, startTime='2020-09-14', endTime='2020-09-14', ktype='1Min')

# 可同时运行多个同类型（多个历史回测或多个实时模拟）回测，registAction的时候注意填入contract_tag参数区分不同合约的回调函数
# 注意：暂不支持同一个StrategiesManager实例历史回测跟实时模拟同时使用
# s_manager.runHistory(stock_code='COMEX_F_GC_2012', startTime='2020-09-10', endTime='2020-09-10', ktype='1Min')



import json
import time

from demo.utils_redis import RedisUtils
from framework.strategies_manager import StrategiesManager

"""
####################################################
# 策略管理器使用例子

实时行情格式：
CFFEX,F,IF,2009,0,2020-09-04 13:51:00,4737.4,4737.4,4735.6,4735.6,0.0,118.0,0,0.0,0.0,0.0,0

0 exchangeNo
1 commodity_type_
2 commodityNo
3 contractNo
4 klType
5 dateTimeStamp
6 highPrice
7 openPrice
8 lowPrice
9 closePrice
10 lastClosePrice
11 volume
12 turnover
13 turnoverRate
14 pe
15 changeRate
16 timestamp

历史数据格式
{'time_key': '2020-07-21 14:53:00', 'open': 4659.2, 'high': 4659.8, 'low': 4657.4, 'close': 4658.8, 'volume': 326.0, 'code': 'CFFEX.IF2008', 'pe_ratio': 0, 'turnover_rate': 0, 'turnover': 0, 'last_close': 4659.2, 'change_rate': -0.0085851648}

####################################################
"""

ri = RedisUtils('./redis.conf')
ri.lookup_redist_info()

# 创建实例，参数可以不传
manager = StrategiesManager(configPath='./sconfig.conf', tempPath='./temp', isShowQuoteMsg=True)


def myAction01(data):
    """
    自定义函数
    :param data: 返回的行情数据，实时返回的是dict类型，历史返回的是逗号分隔的字符串
    :return:
    """
    print(f'_______> {data}')
    time.sleep(2)


def myAction02(data):
    """
    自定义操作
    :param data:
    实时行情数据字符串
    历史行情数据，字典类型
    :return:
    """
    print(f'&&&&&&> {data}')
    # rsp = manager.getHistory(stock_code='NYMEX_F_CL_2010', startTime='2020-08-17', endTime='2020-08-17', ktype='1Min')
    # print(rsp)
    # time.sleep(5)


def myAction03(data):
    print(f'*********> {data}')
    global ri
    if type(data) == dict:
        ri.push_list_value('test2', json.dumps(data, ensure_ascii=False))
    else:
        ri.push_list_value('test2', data)


# 注册自定义函数
manager.registAction(myAction01)
manager.registAction(myAction02)
manager.registAction(myAction03)

# 实时数据使用，支持内外盘合约
manager.runRealTime(stock_code='COMEX_F_GC_2012', ktype='1Min', isGetAllQuote=False)
# 历史数据回测使用，支持内外盘合约
# manager.runHistory(stock_code='NYMEX_F_CL_2010', startTime='2020-08-07', endTime='2020-08-17', ktype='1Min')

# manager.runRealTime(stock_code='CFFEX_F_IF_2009', ktype='1Min', getAllQuote=False)
# manager.runHistory(stock_code='CFFEX_F_IF_2009', startTime='2020-06-07', endTime='2020-08-17', ktype='1Min')



# coding: utf-8
import json
import redis


def file2dict(path):
    """
    从文件读取配置信息转字典
    :param path: 配置文件路径
    :return:
    """
    # 读取配置文件
    with open(path) as f:
        return json.load(f)


class RedisUtils(object):
    """Redis操作类"""

    def __init__(self, conf):
        self.config = file2dict(conf)
        if self.config['passwd'] is None:
            self.r = redis.Redis(host=self.config['redisip'], port=self.config['redisport'], db=self.config['db'])
        else:
            self.r = redis.Redis(host=self.config['redisip'], port=self.config['redisport'], db=self.config['db'],
                                 password=self.config['passwd'])

    def cleanup(self):
        # 清理Redis当前数据库
        self.r.flushdb()

    def lookup_redist_info(self):
        # 查询Redis配置
        info = self.r.info()

    def set_key_value(self, key, value):
        # 设置键值对key<-->value
        self.r.set(key, value)

    def get_key_value(self, key):
        # 查询键值对
        return self.r.get(key)

    def set_hkey_value(self, name, key, value):
        # 设置哈希键值对key<-->value
        self.r.hset(name, key, value)

    def get_hkey_value(self, name, key):
        # 设置哈希键值对key<-->value
        return self.r.hget(name, key)

    def save(self):
        # 强行保存数据到硬盘
        return self.r.save()

    def get_keys(self):
        # 获取当前数据库里面所有键值
        return self.r.keys()

    def delete_key(self, key):
        # 删除某个键
        return self.r.delete(key)

    def push_list_value(self, listname, value):
        # 推入到队列
        return self.r.lpush(listname, value)

    def pull_list_range(self, listname, starpos=0, endpos=-1):
        # 获取队列某个连续片段
        return self.r.lrange(listname, starpos, endpos)

    def get_list_len(self, listname):
        # 获取队列长度
        return self.r.llen(listname)


def main():
    ri = RedisUtils('./redis.conf')
    ri.lookup_redist_info()
    ri.set_key_value('test1', 1)
    ri.push_list_value('test2', 1)
    ri.push_list_value('test2', 2)


if __name__ == '__main__':
    main()



import json
import os
import traceback
from queue import Empty, Queue
from threading import Thread
import zmq
from framework.engine_event import EventEngine, klEvent
from framework.log_handler import DefaultLogHandler
from framework.trader.trade_manager import TradeManager
from framework.utils import file2dict, select_pan_type, isKtypeRight
from framework.utils_requests import get_page_json


class ActionFun:
    """
    用户自定义函数封装
    """
    typeName = ''  # 函数名
    fun = None  # 执行的函数
    tag = ''  # 对应的合约
    threadId = None  # 执行线程


class StrategiesManager:
    _remote_history_url = 'http://{}/gethistory?code={}&startTime={}&endTime={}&ktype={}'
    _threads = {}
    _finished = False
    _action = []
    _tempDir = './temp'
    _prepareActionQueue = Queue()
    _getAllQuote = False
    _quote_temp = {}  # 行情缓存数据
    _max_quote_size = 500  # 行情缓存数据最大条数
    _zmq_clients = {}  # zmq连接

    def __init__(self, configPath='./sconfig.conf', tempPath='./temp', isShowQuoteMsg=True):
        """
        初始化

        配置文件格式：
        {
            "remote_history_url": "127.0.0.1:10039",
            "kl_real_time_addr_in": "tcp://127.0.0.1:10019",
            "kl_real_time_addr_out": "tcp://127.0.0.1:10020"
        }
        :param configPath: 配置文件路径，默认./sconfig.conf
        :param tempPath: 数据缓存文件夹路径，默认./temp
        :param isShowQuoteMsg: 是否输出行情数据，默认True
        """
        self._logHandler = DefaultLogHandler(filename='strategies.log')
        self._logHandler.info('************** 策略管理器初始化 ***************')
        self._config = file2dict(configPath)
        self._url_ip = self._config['remote_history_url']
        self._kl_real_time_addr_in = self._config['kl_real_time_addr_in']
        self._kl_real_time_addr_out = self._config['kl_real_time_addr_out']
        self._tempDir = tempPath
        self._isShowQuoteMsg = isShowQuoteMsg

        self._eventloop = EventEngine(self._logHandler)
        self._eventloop.start()

        self._trade_m = TradeManager(event_loop=self._eventloop)

    def stop(self):
        self._logHandler.debug('---> StrategiesManager stopping...')
        self._finished = True

        for k in self._threads.keys():
            try:
                self._threads[k].join()
            except Exception as e:
                self._logHandler.error(e.args)
                self._logHandler.error(traceback.print_exc())
        self._threads.clear()

        for k in self._zmq_clients.keys():
            self._zmq_clients[k].close()
        self._eventloop.stop()
        self._logHandler.debug('---> StrategiesManager stop finished.')

    def getTradeManager(self):
        """
        获取交易管理器
        :return:
        """
        return self._trade_m

    def getEventloop(self):
        """
        获取事件堆
        :return:
        """
        return self._eventloop

    def registAction(self, func, contract_tag: str):
        """
        注册回调方法
        :param func:回调函数
        :param contract_tag:标记触发的合约，例如：NYMEX_F_CL_2009
        :return:
        """
        af = ActionFun()
        af.typeName = func.__name__
        af.tag = contract_tag
        af.fun = func
        self._action.append(af)

    def getHistory(self, stock_code, startTime='', endTime='', ktype='') -> dict:
        """
        获取合约历史数据
        :param stock_code:
        :param startTime:
        :param endTime:
        :param ktype:
        :return:
        """
        url = self._remote_history_url.format(self._url_ip, stock_code, startTime, endTime, ktype)
        rsp = get_page_json(url)
        return rsp

    def _doHistoryAction(self, stock_code, startTime='', endTime='', ktype=''):
        self._logHandler.info(f'开启任务 ---> 历史回测 {stock_code}合约 从{startTime}到{endTime} {ktype} 数据')
        self._logHandler.info('正在加载数据，请稍后...')

        if not isKtypeRight(ktype):
            self._logHandler.error(f'不支持的类型：{ktype}')
            return

        if not os.path.exists(self._tempDir):
            os.makedirs(self._tempDir)

        start = startTime.replace('-', '')
        end = endTime.replace('-', '')
        filename = f'{stock_code}-{start}-{end}-{ktype}'
        filePath = f'{self._tempDir}/{filename}.dat'
        data_list = None
        # 优先加载本地缓存
        if os.path.exists(filePath):
            data_list = file2dict(filePath)
        # 本地没有缓存从网络获取
        if data_list is None:
            url = self._remote_history_url.format(self._url_ip, stock_code, startTime, endTime, ktype)
            rsp = get_page_json(url)
            if len(rsp) == 0:
                self._logHandler.error(f'获取 {stock_code} 合约行情出错!!')
                return
            data_list = rsp['data']
            if len(data_list) == 0:
                self._logHandler.error(f'从网络获取历史数据返回为空,历史回测 {stock_code}合约 从{startTime}到{endTime} {ktype} 数据')
                return

            # 转统一的合约格式
            for item in data_list:
                item['code'] = stock_code

            # 缓存到本地
            with open(filePath, 'w') as file:
                bJson = json.dumps(data_list, ensure_ascii=False)  # dict转json
                file.writelines(bJson)
                file.flush()

        self._logHandler.info('加载数据完成，执行回测...')
        for item in data_list:
            self._eventloop.put(klEvent(theme=f'{stock_code}-{ktype}', data=item))
            # 缓存内存
            self._save_quote_temp(item['code'].upper(), item)

            # 执行回调函数
            for act in self._action:
                if act.tag == stock_code:
                    act.fun(item)

        key = stock_code + startTime + endTime + ktype
        self._threads.pop(key)
        self._trade_m.save_data()
        self._logHandler.info(f'完成任务 {filename} --->')
        if len(self._threads.keys()) == 0:
            self.stop()

    def _connect_zmq(self, stock_code, ktype):
        try:
            quote_ctx = zmq.Context()
            client = quote_ctx.socket(zmq.SUB)
            # client.setsockopt(zmq.ZMQ_RECONNECT_IVL, 500)
            # client.setsockopt(zmq.ZMQ_RECONNECT_IVL_MAX, 5000)

            pan_type = select_pan_type(stock_code)
            # 请求历史数据
            if pan_type == 'in':
                client.connect(self._kl_real_time_addr_in)
            elif pan_type == 'out':
                client.connect(self._kl_real_time_addr_out)
            else:
                self._logHandler.info("无法判断内外盘")
                return None

            client.setsockopt_string(zmq.SUBSCRIBE, ktype)
            client.setsockopt(zmq.RCVTIMEO, 10000)
            return client
        except zmq.error.ZMQError as e:
            self._logHandler.error("zmq 连接出错:%s" % e)
            return None

    def _doRealTimeAction(self, stock_code, ktype):
        self._logHandler.info(f'开启任务 ---> 实时模拟 {stock_code}合约 {ktype} 数据')
        client = self._connect_zmq(stock_code, ktype)

        key = f'{stock_code}-{ktype}'
        if self._zmq_clients.get(key) is not None:
            try:
                self._zmq_clients.get(key).close()
            except Exception as e:
                self._logHandler.error(e.args)
                self._logHandler.error(traceback.print_exc())
        self._zmq_clients[key] = client

        for action in self._action:
            if action.tag == stock_code:
                self._prepareActionQueue.put(action)

        while not self._finished:
            try:
                if client is not None:
                    response = client.recv()
                else:
                    self._logHandler.error(f'zmq尝试重连......')
                    client = self._connect_zmq(stock_code, ktype)
                    continue
            except zmq.ZMQError as e:
                self._logHandler.error(f'zmq {e.args}')

                client.close()
                self._logHandler.error(f'zmq尝试重连......')
                client = self._connect_zmq(stock_code, ktype)
                continue

            response = str(response, encoding='GB2312')
            if response != ktype:
                if self._isShowQuoteMsg:
                    self._logHandler.info("quote >:  {}".format(response))

                response_list = response.split(',')
                rst = {'time_key': response_list[5], 'open': response_list[7], 'high': response_list[6],
                       'low': response_list[8], 'close': response_list[9], 'volume': response_list[11],
                       'code': f'{response_list[0]}_{response_list[1]}_{response_list[2]}_{response_list[3]}',
                       'pe_ratio': response_list[14], 'turnover_rate': response_list[13], 'turnover': response_list[12],
                       'last_close': response_list[10], 'change_rate': response_list[15]}

                self._eventloop.put(klEvent(theme=f'{stock_code}-{ktype}', data=rst))
                # 缓存
                self._save_quote_temp(rst['code'].upper(), rst)

                if self._getAllQuote \
                        or (rst['code'] == stock_code):
                    size = self._prepareActionQueue.qsize()
                    while size != 0:
                        size -= 1
                        if not self._prepareActionQueue.empty():
                            try:
                                doact = self._prepareActionQueue.get(block=False)
                                if doact is not None:
                                    handle_thread = Thread(target=self._wraperFun, name=doact.typeName,
                                                           args=(doact, rst))
                                    doact.threadId = handle_thread
                                    handle_thread.start()
                            except Empty as e:
                                self._logHandler.error(f'{e.args}')
        # 结束循环，销毁订阅
        if self._zmq_clients.get(key) is not None:
            try:
                self._zmq_clients.get(key).close()
            except Exception as e:
                self._logHandler.error(e.args)
                self._logHandler.error(traceback.print_exc())
            self._zmq_clients.pop(key)

        self._threads.pop(key)
        self._trade_m.save_data()
        self._logHandler.info(f'完成任务 {key}--->')
        if len(self._threads.keys()) == 0:
            self.stop()

    def _wraperFun(self, doactfun, data):
        if isinstance(doactfun, ActionFun):
            try:
                doactfun.fun(data)
            except Exception as e:
                self._logHandler.error(f'执行任务<{doactfun.typeName}>发生异常：{e.args}')
                self._logHandler.error(traceback.print_exc())
            doactfun.threadId = None
            self._prepareActionQueue.put(doactfun)

    def _save_quote_temp(self, contract_name: str, data):
        """
        缓存行情数据
        :param contract_name:合约名，例如：NYMEX_F_IF_2009
        :param data:行情数据
        :return:
        """
        quote_list = self._quote_temp.get(contract_name)
        if quote_list is None:
            self._quote_temp[contract_name] = [data]
        else:
            last_data = quote_list[-1]
            if last_data['time_key'] == data['time_key']:
                # del quote_list[-1]
                # quote_list.append(data)
                quote_list[-1] = data
                return

            if len(quote_list) > self._max_quote_size:
                del quote_list[0]
            quote_list.append(data)
            # self._logHandler.debug(f'合约{contract_name}当前缓存条数：{len(quote_list)}')

    def get_history_quote(self, contract_name: str):
        """
        获取行情缓存数据列表
        :param contract_name: 合约名，例如：NYMEX_F_IF_2009
        :return:
        """
        quote_list = self._quote_temp.get(contract_name)
        # self._logHandler.debug(f'合约{contract_name}当前缓存条数：{len(quote_list)}')
        return quote_list

    def runRealTime(self, stock_code, ktype='1Min', isGetAllQuote=False):
        """
        运行实时数据模拟
        :param stock_code:例如：COMEX_F_GC_2012
        :param ktype: k线类型，默认1Min，例如：1Min
        :param isGetAllQuote:是否获取所有合约行情
        :return:
        """
        if stock_code == "":
            self._logHandler.error('合约信息为空，请输入合约代码')
            return
        self._getAllQuote = isGetAllQuote
        key = stock_code + ktype
        if self._threads.get(key) is None:
            self._threads[key] = Thread(target=self._doRealTimeAction, args=[stock_code, ktype], name="runRealTime")
            self._threads[key].start()
        else:
            self._logHandler.info('任务已存在 --->')

    def runHistory(self, stock_code, startTime='', endTime='', ktype='1Min'):
        """
        运行历史回测
        :param stock_code:
        :param startTime:
        :param endTime:
        :param ktype:
        :return:
        """
        key = stock_code + startTime + endTime + ktype
        if self._threads.get(key) is None:
            self._threads[key] = Thread(target=self._doHistoryAction, args=[stock_code, startTime, endTime, ktype],
                                        name="runHistory")
            self._threads[key].start()
        else:
            self._logHandler.info('任务已存在 --->')



import json
import time
import traceback
from threading import Thread

import zmq

from framework.engine_event import EventEngine, klEvent
from framework.log_handler import DefaultLogHandler
from framework.utils import file2dict, select_pan_type, isKtypeRight
from framework.utils_requests import get_page_json


class QuoteManager:
    _remote_history_url = 'http://{}/gethistory?code={}&startTime={}&endTime={}&ktype={}'
    _finished = False
    _threads = {}
    _zmq_clients = {}
    _quote_temp = {}  # 行情缓存数据
    _max_quote_size = 500  # 行情缓存数据最大条数

    def __init__(self, event_loop: EventEngine = None, configPath='./sconfig.conf'
                 , tempPath='./temp', isShowQuoteMsg=True):
        self._getAllQuote = False
        self._config = file2dict(configPath)
        self._url_ip = self._config['remote_history_url']
        self._kl_real_time_addr_in = self._config['kl_real_time_addr_in']
        self._kl_real_time_addr_out = self._config['kl_real_time_addr_out']
        self._tempDir = tempPath
        self._isShowQuoteMsg = isShowQuoteMsg
        self._logHandler = DefaultLogHandler(filename='QuoteManager.log')
        self._logHandler.info('************** 行情管理器初始化 ***************')

        if event_loop is None:
            self._event_loop = EventEngine(self._logHandler)
        else:
            self._event_loop = event_loop
        self._event_loop.start()

    def getEventloop(self):
        """
        获取事件堆
        :return:
        """
        return self._event_loop

    def stop(self):
        self._finished = True
        for k in self._threads.keys():
            self._threads[k].join()
        self._threads.clear()
        for k in self._zmq_clients.keys():
            self._zmq_clients[k].close()

        self._event_loop.put(klEvent(theme='exit-loop', data=''))  # 发出停止信号
        while self._event_loop.isActivate():
            time.sleep(3)

        self._event_loop.stop()
        self._logHandler.debug('QuoteManager stop finished.')

    def registQuoteListener(self, fun_listener, key: str):
        if fun_listener is not None:
            self._event_loop.register(key, fun_listener)

    def getHistory(self, stock_code, startTime='', endTime='', ktype='', fun_listener=None):
        """
        获取合约历史数据
        :param stock_code:
        :param startTime:
        :param endTime:
        :param ktype:
        :param fun_listener: 行情数据回调函数
        :return:
        """
        if not isKtypeRight(ktype):
            self._logHandler.error(f'不支持的类型：{ktype}')
            return

        start = startTime.replace('-', '')
        end = endTime.replace('-', '')
        #
        import os
        if not os.path.exists(self._tempDir):
            os.makedirs(self._tempDir)
        #
        key = f'{stock_code}-{start}-{end}-{ktype}'
        filePath = f'{self._tempDir}/{key}.dat'
        data_list = None
        if os.path.exists(filePath):
            data_list = file2dict(filePath)
        #
        if data_list is None:
            url = self._remote_history_url.format(self._url_ip, stock_code, startTime, endTime, ktype)
            rsp = get_page_json(url)
            data_list = rsp['data']
            if len(data_list) == 0:
                self._logHandler.error(f'从网络获取历史数据返回为空,历史回测 {stock_code}合约 从{startTime}到{endTime} {ktype} 数据')
                return None

            # 转统一的合约格式
            for item in data_list:
                item['code'] = stock_code

            # 缓存到本地
            with open(filePath, 'w') as file:
                bJson = json.dumps(data_list, ensure_ascii=False)  # dict转json
                file.writelines(bJson)
                file.flush()

        # 注册事件监听回调函数
        self.registQuoteListener(fun_listener, key)

        if fun_listener is not None:
            if not self._event_loop.isActivate():
                self._event_loop.start()
            for item in data_list:
                self._event_loop.put(klEvent(theme=key, data=item, is_async=False))
        # for temp in data_list:
        #     self.save_quote_temp(stock_code, temp)
        return data_list

    def getRealTimeKL(self, stock_code: str, ktype: str, fun_listener, isGetAllQuote=False):
        """
        获取实时数据
        :param stock_code:
        :param ktype:
        :param fun_listener:
        :param isGetAllQuote: 是否获取所有订阅行情,默认False
        :return:
        """
        if not isKtypeRight(ktype):
            self._logHandler.error(f'不支持的类型：{ktype}')
            return

        key = f'{stock_code}-{ktype}'
        # 注册事件监听回调函数
        self.registQuoteListener(fun_listener, key)
        self._getAllQuote = isGetAllQuote

        if self._threads.get(key) is None:
            _watch_thread = Thread(target=self._doRealTimeAction, args=[stock_code, ktype],
                                   name=f'runRealTime-{stock_code}-{ktype}')
            self._threads[key] = _watch_thread
            _watch_thread.start()
        else:
            self._logHandler.error(f'已订阅合约：{stock_code}-{ktype}')

    def save_quote_temp(self, contract_name: str, data):
        """
        缓存行情数据
        :param contract_name:合约名，例如：NYMEX_F_IF_2009
        :param data:行情数据
        :return:
        """
        quote_list = self._quote_temp.get(contract_name)
        if quote_list is None:
            self._quote_temp[contract_name] = [data]
        else:
            last_data = quote_list[-1]
            # ll = str(last_data).split(',')
            # ll2 = data.split(',')
            # if ll[5] == ll2[5]:
            #     del quote_list[-1]
            #     quote_list.append(data)
            #     return

            if last_data['time_key'] == data['time_key']:
                quote_list[-1] = data
                return

            if len(quote_list) > self._max_quote_size:
                del quote_list[0]
            quote_list.append(data)

    def get_quote_temp(self, contract_name: str):
        """
        获取行情缓存数据列表
        :param contract_name: 合约名，例如：NYMEX_F_IF_2009
        :return:
        """
        quote_list = self._quote_temp.get(contract_name)
        return quote_list

    def _connect_zmq(self, stock_code, ktype):
        try:
            quote_ctx = zmq.Context()
            client = quote_ctx.socket(zmq.SUB)
            # client.setsockopt(zmq.ZMQ_RECONNECT_IVL, 500)
            # client.setsockopt(zmq.ZMQ_RECONNECT_IVL_MAX, 5000)

            pan_type = select_pan_type(stock_code)
            # 请求历史数据
            if pan_type == 'in':
                client.connect(self._kl_real_time_addr_in)
            elif pan_type == 'out':
                client.connect(self._kl_real_time_addr_out)
            else:
                self._logHandler.info("无法判断内外盘")
                return None

            client.setsockopt_string(zmq.SUBSCRIBE, ktype)
            client.setsockopt(zmq.RCVTIMEO, 10000)
            return client
        except zmq.error.ZMQError as e:
            self._logHandler.error("zmq 连接出错:%s" % e)
            return None

    def _doRealTimeAction(self, stock_code, ktype):
        self._logHandler.info(f'开启任务 ---> \n实时模拟 {stock_code}合约 {ktype} 数据')
        client = self._connect_zmq(stock_code, ktype)
        key = f'{stock_code}-{ktype}'
        if self._zmq_clients.get(key) is not None:
            try:
                self._zmq_clients.get(key).close()
            except Exception as e:
                self._logHandler.error(e.args)
                self._logHandler.error(traceback.print_exc())

        self._zmq_clients[key] = client

        while not self._finished:
            try:
                if client is not None:
                    response = client.recv()
                else:
                    self._logHandler.error(f'zmq尝试重连......')
                    client = self._connect_zmq(stock_code, ktype)
                    continue
            except zmq.ZMQError as e:
                self._logHandler.error(f'zmq {e.args}')

                client.close()
                self._logHandler.error(f'zmq尝试重连......')
                client = self._connect_zmq(stock_code, ktype)
                continue

            response = str(response, encoding='GB2312')
            if response != ktype:
                if self._isShowQuoteMsg:
                    self._logHandler.debug(response)
                response_list = response.split(',')
                rst = {}
                rst['time_key'] = response_list[5]
                rst['open'] = response_list[7]
                rst['high'] = response_list[6]
                rst['low'] = response_list[8]
                rst['close'] = response_list[9]
                rst['volume'] = response_list[11]
                rst['code'] = f'{response_list[0]}_{response_list[1]}_{response_list[2]}_{response_list[3]}'
                rst['pe_ratio'] = response_list[14]
                rst['turnover_rate'] = response_list[13]
                rst['turnover'] = response_list[12]
                rst['last_close'] = response_list[10]
                rst['change_rate'] = response_list[15]

                if self._getAllQuote \
                        or (rst['code'] == stock_code):
                    self._event_loop.put(klEvent(theme=key, data=response))

                # 缓存
                self.save_quote_temp(rst['code'].upper(), rst)

        # 结束循环，销毁订阅
        if self._zmq_clients.get(key) is not None:
            try:
                self._zmq_clients.get(key).close()
            except Exception as e:
                self._logHandler.error(e.args)
                self._logHandler.error(traceback.print_exc())
            self._zmq_clients.pop(key)
        self._logHandler.info(f'完成 {stock_code}合约 {ktype} 数据实时模拟任务')
